/**
 * License Agreement.
 *
 *  JBoss RichFaces 3.0 - Ajax4jsf Component Library
 *
 * Copyright (C) 2007  Exadel, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 */

package org.richfaces.component.state;

import java.io.IOException;
import java.util.HashSet;
import java.util.Set;

import javax.faces.application.FacesMessage;
import javax.faces.component.StateHolder;
import javax.faces.component.UIComponentBase;
import javax.faces.context.FacesContext;
import javax.faces.el.ValueBinding;

import org.ajax4jsf.ajax.repeat.DataComponentState;
import org.ajax4jsf.ajax.repeat.DataVisitor;
import org.ajax4jsf.ajax.repeat.Range;
import org.richfaces.component.CacheableTreeDataModel;
import org.richfaces.component.TreeRowKey;
import org.richfaces.component.UITree;
import org.richfaces.component.preserve.TreeDataBatchLocator;
import org.richfaces.component.preserve.TreeDataLocator;
import org.richfaces.component.state.events.TreeStateCommandsListener;

/**
 * @author Nick Belaevski - nbelaevski@exadel.com created 23.11.2006
 * 
 */
public class TreeState implements DataComponentState, TreeStateCommandsListener,StateHolder {
	private final class Visitor implements DataVisitor {
		private TreeRowKey key;

		public Visitor(TreeRowKey key) {
			super();
			this.key = key;
		}

		public void process(FacesContext context, Object rowKey, Object argument)
				throws IOException {
			tree.setRowKey(rowKey);
			if (tree.isRowAvailable()) {
				TreeRowKey nextKey = (TreeRowKey) rowKey;

				if (!tree.isLeaf() && nextKey != null) {
					if (key == null || nextKey.isSubKey(key)
							|| nextKey.equals(key)) {
						if (tree.isImmediate()) {
							queuedExpandedNodes.add(nextKey);
						} else {
							expandedNodes.add(nextKey);
						}
						queuedCollapsedNodes.remove(nextKey);
					}
				}
			} else {
				FacesMessage message = new FacesMessage("Row key: " + rowKey
						+ " isn't available!");
				message.setSeverity(FacesMessage.SEVERITY_ERROR);

				context.addMessage(tree.getBaseClientId(context), message);
			}
		}
	}

	private final static TreeRange RANGE_UNCONSTRAINED = new TreeRange() {
		public boolean processChildren(TreeRowKey rowKey) {
			return true;
		}

		public boolean processNode(TreeRowKey rowKey) {
			return true;
		}
	};

	private TreeRowKey selectedNode = null;

	private Set expandedNodes = new HashSet();

	private Set queuedExpandedNodes = new HashSet();

	private Set queuedCollapsedNodes = new HashSet();

	private UITree tree;
	

	public TreeState(UITree tree, boolean stopInCollapsed) {
		super();
		this.stopInCollapsed = stopInCollapsed;
		this.tree = tree;
	}

	public boolean isExpanded(TreeRowKey rowKey) {
		if (rowKey == null) {
			return true;
		}

		return expandedNodes.contains(rowKey)
				|| queuedExpandedNodes.contains(rowKey);
	}

	public boolean isSelected(TreeRowKey rowKey) {
		return (selectedNode != null && selectedNode.equals(rowKey));
	}

	public boolean setSelected(TreeRowKey rowKey) {
		if (selectedNode == null) {
			if (rowKey != null) {
				selectedNode = rowKey;

				return true;
			}

			return false;
		} else {
			if (rowKey != null) {
				if (!selectedNode.equals(rowKey)) {
					selectedNode = rowKey;

					return true;
				}

				return false;
			} else {
				selectedNode = null;

				return true;
			}
		}
	}

	private boolean _transient;

	private boolean stopInCollapsed;

	private transient TreeRange treeRange = null;

	public Range getRange() {
		if (treeRange != null) {
			return treeRange;
		}

		if (stopInCollapsed) {
			return new TreeRange() {

				public boolean processChildren(TreeRowKey rowKey) {
					if (rowKey == null) {
						return true;
					}

					return expandedNodes.contains(rowKey);
				}

				public boolean processNode(TreeRowKey rowKey) {
					return true;
				}
			};
		} else {
			return RANGE_UNCONSTRAINED;
		}
	}

	public boolean isTransient() {
		return _transient;
	}

	public void restoreState(FacesContext context, Object state) {
		Object[] _state = (Object[]) state;
		expandedNodes = (Set) _state[0];
		_transient = ((Boolean) _state[1]).booleanValue();
		stopInCollapsed = ((Boolean) _state[2]).booleanValue();
		selectedNode = (TreeRowKey) _state[3];
		queuedExpandedNodes = (Set) _state[4];
		queuedCollapsedNodes = (Set) _state[5];
		treeDataModel = (CacheableTreeDataModel) UIComponentBase.restoreAttachedState(context, _state[6]);
		ValueBinding treeDataLocatorBinding = tree
				.getValueBinding(UITree.TREE_DATA_LOCATOR_BINDING);
		if (treeDataLocatorBinding != null) {
			Object object = treeDataLocatorBinding.getValue(context);
			if (object instanceof TreeDataLocator) {
				final TreeDataLocator treeDataLocator = (TreeDataLocator) object;
				if (treeDataModel != null) {
					try {
						treeDataModel.walk(context, new DataVisitor() {

							public void process(FacesContext arg0, Object arg1,
									Object arg2) throws IOException {
								treeDataModel.setRowKey(arg1);
								treeDataModel.setRowData(treeDataLocator
										.createData(treeDataModel.getRowData(),
												(TreeRowKey) arg1));
							}

						}, null, null);
					} catch (IOException e) {
						e.printStackTrace();
					}
				}
			} else {
				TreeDataBatchLocator treeDataBatchLocator = (TreeDataBatchLocator) object;
				// TODO implement batch processing
			}
		}
	}

	public Object saveState(FacesContext context) {
		Object[] state = new Object[7];
		state[0] = expandedNodes;
		state[1] = new Boolean(_transient);
		state[2] = new Boolean(stopInCollapsed);
		state[3] = selectedNode;
		state[4] = queuedExpandedNodes;
		state[5] = queuedCollapsedNodes;
		if (!UITree.PRESERVE_MODEL_REQUEST.equals(tree.getPreserveModel())) {
			ValueBinding treeDataLocatorBinding = tree
					.getValueBinding(UITree.TREE_DATA_LOCATOR_BINDING);
			if (treeDataLocatorBinding != null) {
				Object object = treeDataLocatorBinding.getValue(context);
				if (object instanceof TreeDataLocator) {
					final TreeDataLocator treeDataLocator = (TreeDataLocator) object;
					if (treeDataModel != null) {
						try {
							treeDataModel.walk(context, new DataVisitor() {

								public void process(FacesContext arg0,
										Object arg1, Object arg2)
										throws IOException {
									treeDataModel.setRowKey(arg1);
									treeDataModel.setRowData(treeDataLocator
											.createLocator(treeDataModel
													.getRowData(),
													(TreeRowKey) arg1));
								}

							}, null, null);
						} catch (IOException e) {
							e.printStackTrace();
						}
					}
				} else {
					TreeDataBatchLocator treeDataBatchLocator = (TreeDataBatchLocator) object;
					// TODO implement batch processing
				}
			}
			state[6] = UIComponentBase.saveAttachedState(context, treeDataModel);
		}
		return state;
	}

	public void setTransient(boolean newTransientValue) {
		this._transient = newTransientValue;
	}

	public boolean isStopInCollapsed() {
		return stopInCollapsed;
	}

	public void setStopInCollapsed(boolean stopInCollapsed) {
		this.stopInCollapsed = stopInCollapsed;
	}

	private void visitNodes(TreeRange treeRange, TreeRowKey rootKey)
			throws IOException {
		try {
			this.treeRange = treeRange;
			Object oldKey = tree.getRowKey();

			tree.walkModel(FacesContext.getCurrentInstance(), new Visitor(
					rootKey), rootKey, null);
			tree.setRowKey(oldKey);
		} finally {
			this.treeRange = null;
		}
	}

	public void expandAll() throws IOException {
		FacesContext context = FacesContext.getCurrentInstance();
		queuedCollapsedNodes.clear();
		// SubTreeChildrenAppender infoAppender = null;
		//
		// if (storedPersister != null) {
		// storedPersister.reset();
		// infoAppender = storedPersister.getAppender(null);
		// }

		visitNodes(RANGE_UNCONSTRAINED, null);
	}

	public void collapseAll() throws IOException {
		expandedNodes.clear();
		queuedExpandedNodes.clear();
		// RequestUtils.setStoredPersister(tree.getBaseClientId(context),
		// context, null);
	}

	public void collapseNode(TreeRowKey rowKey) throws IOException {
		expandedNodes.remove(rowKey);
		queuedExpandedNodes.remove(rowKey);
		queuedCollapsedNodes.add(rowKey);
		// TreePersister storedPersister =
		// RequestUtils.getStoredPersister(tree.getBaseClientId(context),
		// context);
		// if (storedPersister != null) {
		// storedPersister.removeNode(rowKey);
		// }
	}

	public void expandNode(final TreeRowKey rowKey) throws IOException {
		FacesContext context = FacesContext.getCurrentInstance();
		// SubTreeChildrenAppender infoAppender;
		// if (storedPersister != null) {
		// infoAppender = storedPersister.getAppender(rowKey);
		// } else {
		// infoAppender = null;
		// }

		TreeRange range;
		if (treeDataModel == null) {
			// it's enough to traverse only subkeys of the node
			// we're opening
			range = new TreeRange() {
				public boolean processChildren(TreeRowKey nextKey) {
					return true;
				}

				public boolean processNode(TreeRowKey nextKey) {
					return (rowKey == null && nextKey == null)
							|| nextKey.equals(rowKey)
							|| nextKey.isSubKey(rowKey);
				}

			};
			visitNodes(range, rowKey);
		} else {
			final TreeRange expandedRange = (TreeRange) getRange();
			range = new TreeRange() {

				public boolean processChildren(TreeRowKey key) {
					return expandedRange.processChildren(key) || key == null
							|| key.isSubKey(rowKey);
				}

				public boolean processNode(TreeRowKey key) {
					return expandedRange.processNode(key) || key == null
							|| key.isSubKey(rowKey) || key.equals(rowKey);
				}

			};
			visitNodes(range, rowKey);
		}

	}

	public void transferQueuedNodes() {
		expandedNodes.addAll(queuedExpandedNodes);
		queuedExpandedNodes.clear();
		expandedNodes.removeAll(queuedCollapsedNodes);
		queuedCollapsedNodes.clear();
	}

	private CacheableTreeDataModel treeDataModel;

	public CacheableTreeDataModel getTreeDataModel() {
		return treeDataModel;
	}

	public void setTreeDataModel(CacheableTreeDataModel treeDataModel) {
		this.treeDataModel = treeDataModel;
	}
}
