/**
 * License Agreement.
 *
 *  JBoss RichFaces 3.0 - Ajax4jsf Component Library
 *
 * Copyright (C) 2007  Exadel, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 */

package org.richfaces.renderkit;

import java.io.IOException;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.faces.FacesException;
import javax.faces.component.NamingContainer;
import javax.faces.component.UIComponent;
import javax.faces.context.FacesContext;
import javax.faces.context.ResponseWriter;

import org.ajax4jsf.ajax.repeat.DataVisitor;
import org.ajax4jsf.framework.renderer.AjaxRendererUtils;
import org.ajax4jsf.framework.renderer.ComponentVariables;
import org.ajax4jsf.framework.renderer.ComponentsVariableResolver;
import org.ajax4jsf.framework.util.javascript.JSFunction;
import org.ajax4jsf.framework.util.javascript.JSReference;
import org.richfaces.component.LastElementAware;
import org.richfaces.component.TreeRowKey;
import org.richfaces.component.UITree;
import org.richfaces.component.UITreeNode;
import org.richfaces.component.nsutils.NSUtils;

public abstract class TreeRendererBase extends CompositeRenderer {

	private class DataVisitorWithLastElement implements DataVisitor,
	LastElementAware {

		private boolean isLastElement = false;

		private final ResponseWriter writer;

		private final StringHolder holder;

		private final Flag flag;

		private final UITree tree;

		private final TreeDataModelEventNavigator navigator;

		private DataVisitorWithLastElement(ResponseWriter writer,
				StringHolder holder, Flag flag, UITree tree,
				TreeDataModelEventNavigator navigator) {
			this.writer = writer;
			this.holder = holder;
			this.flag = flag;
			this.tree = tree;
			this.navigator = navigator;
		}

		public void process(FacesContext context, Object rowKey, Object argument)
		throws IOException {

			navigator.followRowKey((TreeRowKey) rowKey);

			if (tree.isSelected()) {
				holder.setValue(tree.getClientId(context));
			}

			if (!flag.isFlagSet()) {
				closeLevelDownTable(context, tree, writer);
				openLevelDownTable(context, tree, writer);
			} else {
				flag.resetFlag();
			}

			UITreeNode nodeFacet = tree.getNodeFacet();
			Object oldAttrValue = nodeFacet.getAttributes().get("isLastElement");
			try {
				nodeFacet.getAttributes().put("isLastElement", new Boolean(isLastElement));

				renderChild(context, nodeFacet);
			} finally {
				if (oldAttrValue != null) {
					nodeFacet.getAttributes().put("isLastElement", oldAttrValue);
				} else {
					nodeFacet.getAttributes().remove("isLastElement");
				}
			}
		}

		public void setLastElement() {
			isLastElement = true;
			navigator.setLastElement();
		}

		public void resetLastElement() {
			isLastElement = false;
			navigator.resetLastElement();
		}
	}

	public TreeRendererBase() {
		super();
		addContributor(DraggableRendererContributor.getInstance());
		addContributor(DropzoneRendererContributor.getInstance());
	
		addParameterEncoder(DnDParametersEncoder.getInstance());
	}

	public void writeNamespace(FacesContext context, UIComponent component) throws IOException {
		NSUtils.writeNameSpace(context, component);
	}

	public void encodeAjaxChild(FacesContext context, UIComponent component,
			String path, Set ids, Set renderedAreas) throws IOException {
		super.encodeAjaxChild(context, component, path, ids, renderedAreas);

		try {
			if (component instanceof UITree) {
				UITree tree = (UITree) component;
				tree.captureOrigValue();
				//Object rowKey = tree.getRowKey();

				boolean encodeScripts = false;

				tree.setRowKey(null);

				//we should add xmlns to AJAX response
				//we'll write neutral inner element and add xmlns there
				ResponseWriter responseWriter = context.getResponseWriter();
				responseWriter.startElement("div", tree);
				writeNamespace(context, component);

				try {
					Set ajaxKeys = tree.getAjaxKeys(); 
					if (ajaxKeys != null) {
						Iterator ajaxKeysItr = tree.getAjaxKeys().iterator();
						while (ajaxKeysItr.hasNext()) {
							TreeRowKey key = (TreeRowKey) ajaxKeysItr.next();
							tree.setRowKey(key);

							String id = tree.getClientId(context);
							if (ids.isEmpty() || ids.contains(id)) {
								writeContent(context, tree, key);
								encodeScripts = true;
								renderedAreas.add(id);
							}
						}
					}
				} catch (Exception e) {
					throw new FacesException(e);
				} finally {
					tree.setRowKey(null);
					tree.restoreOrigValue();
				}
				if (encodeScripts) {
					writeScript(context, tree, renderedAreas);
				}

				String inputId = encodeSelectionStateInput(context, tree);
				if (inputId != null) {
					renderedAreas.add(inputId);
				}
				
				responseWriter.endElement("div");
			}
		} finally {
			ComponentsVariableResolver.removeVariables(this, component);
		}
	}

	public String encodeSelectionStateInput(FacesContext context, UITree tree) throws IOException {
		ComponentVariables variables = ComponentsVariableResolver.getVariables(this, tree);
		Object object = variables.getVariable("selectionId");
		if (object != null) {
			ResponseWriter writer = context.getResponseWriter();
			writer.startElement("input", tree);
			writer.writeAttribute("type", "hidden", null);
			String selectionHolderInputId = tree.getSelectionStateInputName(context);
			writer.writeAttribute("id", selectionHolderInputId, null);
			writer.writeAttribute("name", selectionHolderInputId, null);
			
			writer.writeAttribute("value", object,
						null);
			writer.endElement("input");
		
			return selectionHolderInputId;
		}
		
		return null;
	}
	
	protected String getAjaxScript(FacesContext context, UITree tree) {
		String id = tree.getBaseClientId(context);
		JSFunction function = AjaxRendererUtils
		.buildAjaxFunction(tree, context);
		Map eventOptions = AjaxRendererUtils.buildEventOptions(context, tree);
		Map parameters = (Map) eventOptions.get("parameters");
		parameters.remove(id);
		parameters.put(id + UITree.SELECTED_NODE_PARAMETER_NAME,
				new JSReference("event.selectedNode"));
		function.addParameter(eventOptions);
		StringBuffer buffer = new StringBuffer();
		function.appendScript(buffer);
		buffer.append("; return false;");
		return buffer.toString();
	}

	protected String getDragIndicatorId(FacesContext context, UITree tree) {
		if(tree.getAttributes().get("dragIndicator") instanceof String) {
			String dragIndicator = (String) tree.getAttributes().get("dragIndicator");
			if (dragIndicator != null) {
				UIComponent component = tree.getParent().findComponent(dragIndicator);
				if (component != null) {
					return component.getClientId(context);
				}
			}
		}

		return "";
	}

	protected String getWaitOnLoadId(FacesContext context, UITree tree) {
		if(tree.getAttributes().get("waitOnLoadId") instanceof String) {
			String waitOnLoadId = (String) tree.getAttributes().get("waitOnLoadId");
			if (waitOnLoadId != null) {
				UIComponent component = tree.getParent().findComponent(waitOnLoadId);
				if (component != null) {
					return component.getClientId(context);
				}
			}
		}

		return "";
	}

	protected String getScriptContributions(FacesContext context, UITree tree) {
		return super.getScriptContributions(getJavaScriptVarName(context, tree), context, tree);
	}

	protected String getJavaScriptVarName(FacesContext context, UITree tree) {
		String id = tree.getBaseClientId(context);
		return "Richfaces_Tree_" + id.replaceAll("[^A-Za-z0-9_]", "_");
	}

	private void writeScript(FacesContext context, UITree tree,
			Set renderedAreas) throws IOException {
		final ResponseWriter writer = context.getResponseWriter();
		final String clientId = tree.getBaseClientId(context);

		String scriptId = clientId + NamingContainer.SEPARATOR_CHAR + "script";
		writer.startElement("div", tree);
		getUtils().writeAttribute(writer, "id", scriptId);
		writer.startElement("script", tree);
		getUtils().writeAttribute(writer, "type", "text/javascript");

		Set set = tree.getAjaxKeys();
		String varName = getJavaScriptVarName(context, tree);

		if (set != null) {
			Iterator iterator = set.iterator();
			while (iterator.hasNext()) {
				TreeRowKey rowKey = (TreeRowKey) iterator.next();

				writer.writeText(varName + ".getNodeElements(" + varName
						+ ", \"" + clientId + NamingContainer.SEPARATOR_CHAR
						+ rowKey.toString() + "\");", null);
			}

			set.clear();
		}

		writer.endElement("script");
		writer.endElement("div");

		renderedAreas.add(scriptId);

		renderedAreas.add(tree.getClientId(context)
				+ NamingContainer.SEPARATOR_CHAR + "input");
	}

	public void encodeChildren(FacesContext context, UIComponent component)
	throws IOException {

		writeContent(context, (UITree) component, null);
	}

	public void writeContent(final FacesContext context, final UITree input)
	throws IOException {
		writeContent(context, input, null);
	}

	public void writeContent(final FacesContext context, final UITree input,
			TreeRowKey key) throws IOException {
		final ResponseWriter writer = context.getResponseWriter();

		// simple flag can be used here because
		// we cannot jump more than one level down until next node
		// when rendering
		final Flag droppedDownToLevelFlag = new Flag();

		TreeRowKey rowKey = (TreeRowKey) key;
		final TreeDataModelEventNavigator levelNavigator = new TreeDataModelEventNavigator(
				input, context, rowKey) {

			public void afterDown() throws IOException {
				openLevelDownTable(context, input, writer);
			}

			public void afterUp(int levels) throws IOException {
				for (int i = 0; i < levels; i++) {
					closeLevelDownTable(context, input, writer);
					writer.endElement("td");
					writer.endElement("tr");
				}
			}

			public void beforeDown() throws IOException {
				droppedDownToLevelFlag.setFlag();

				String id = getTree().getClientId(context)
				+ NamingContainer.SEPARATOR_CHAR;

				writer.startElement("tr", input);
				getUtils().writeAttribute(writer, "id", id + "childs");

				if (!getTree().isExpanded()) {
					getUtils()
					.writeAttribute(writer, "style", "display: none;");
				}

				writer.startElement("td", input);
				if (!isStackedLastElement()) {
					if (getTree().isShowConnectingLines() && this.getRowKey() != null)
						getUtils().writeAttribute(writer, "class",
						"dr-tree-h-ic-line");
				}
				writer.endElement("td");
				writer.startElement("td", input);
				getUtils().writeAttribute(writer, "colspan", "2");
				getUtils().writeAttribute(writer, "id", id + "childs:td");
				getUtils().writeAttribute(writer, "valign", "top");
			}

			public void beforeUp(int levels) {
			}
		};

		final StringHolder selectionValueHolder = new StringHolder();

		//Object savedRowKey = input.getRowKey();
		try {
			input.captureOrigValue();
			
			if (key != null) {
				droppedDownToLevelFlag.setFlag();
				openLevelDownTable(context, input, writer);
				input.setRowKey((TreeRowKey) key);
			} else {
				input.setRowKey(null);
			}

			input.walk(context, new DataVisitorWithLastElement(writer,
					selectionValueHolder, droppedDownToLevelFlag, input,
					levelNavigator), key, null);

			if (key != null) {
				closeLevelDownTable(context, input, writer);
			}

			levelNavigator.followRowKey(null);

			ComponentVariables variables = ComponentsVariableResolver.getVariables(this, input);
			String value = selectionValueHolder.getValue();
			if (value != null) {
				variables.setVariable("selectionId", value);
			} else if (rowKey == null) {
				variables.setVariable("selectionId", "");
			}
		} finally {
			input.setRowKey(null);
			input.restoreOrigValue();
		}
	}

	private void openLevelDownTable(FacesContext context, UITree tree,
			ResponseWriter writer) throws IOException {
		writer.startElement("table", tree);

		getUtils().writeAttribute(writer, "border", "0");
		getUtils().writeAttribute(writer, "cellpadding", "0");
		getUtils().writeAttribute(writer, "cellspacing", "0");
		getUtils().writeAttribute(writer, "id", tree.getClientId(context));
		getUtils().writeAttribute(writer, "style", "width: 100%;");

		writer.startElement("tbody", tree);
	}

	private void closeLevelDownTable(FacesContext context, UITree tree,
			ResponseWriter writer) throws IOException {
		writer.endElement("tbody");
		writer.endElement("table");
	}
}

class Flag {
	private boolean flag = false;

	public void setFlag() {
		this.flag = true;
	}

	public void resetFlag() {
		this.flag = false;
	}

	public boolean isFlagSet() {
		return flag;
	}
}

class StringHolder {
	private String value;

	public String getValue() {
		return value;
	}

	public void setValue(String value) {
		this.value = value;
	}
}